<?php

use Blesta\Core\Util\Validate\Server;

/**
 * Convoy Module
 *
 * @link https://aleksa.tf Aleksa Đorđić
 * @link https://code-cats.com Code Cats
 */
class Convoy extends Module
{
    /**
     * {@inheritdoc}
     */
    public function __construct()
    {
        // Load the language required by this module
        Language::loadLang('convoy', null, dirname(__FILE__) . DS . 'language' . DS);

        // Load components required by this module
        Loader::loadComponents($this, ['Input']);

        // Load module config
        $this->loadConfig(dirname(__FILE__) . DS . 'config.json');

        Configure::load('convoy', dirname(__FILE__) . DS . 'config' . DS);
    }

    /**
     * {@inheritdoc}
     */
    public function install()
    {
    }

    /**
     * {@inheritdoc}
     */
    public function upgrade($current_version)
    {
    }

    /**
     * {@inheritdoc}
     */
    public function uninstall($module_id, $last_instance)
    {
    }

    /**
     * {@inheritdoc}
     */
    public function manageModule($module, array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('manage', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'convoy' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html', 'Widget']);

        $this->view->set('module', $module);

        return $this->view->fetch();
    }

    /**
     * {@inheritdoc}
     */
    public function manageAddRow(array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('add_row', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'convoy' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html', 'Widget']);

        if (!empty($vars)) {
            // Set unset checkboxes
            $checkbox_fields = [];

            foreach ($checkbox_fields as $checkbox_field) {
                if (!isset($vars[$checkbox_field])) {
                    $vars[$checkbox_field] = 'false';
                }
            }
        }

        $this->view->set('vars', (object) $vars);

        return $this->view->fetch();
    }

    /**
     * {@inheritdoc}
     */
    public function manageEditRow($module_row, array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('edit_row', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'convoy' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html', 'Widget']);

        if (empty($vars)) {
            $vars = $module_row->meta;
        } else {
            // Set unset checkboxes
            $checkbox_fields = [];

            foreach ($checkbox_fields as $checkbox_field) {
                if (!isset($vars[$checkbox_field])) {
                    $vars[$checkbox_field] = 'false';
                }
            }
        }

        $this->view->set('vars', (object) $vars);

        return $this->view->fetch();
    }

    /**
     * {@inheritdoc}
     */
    public function addModuleRow(array &$vars)
    {
        $meta_fields = ['name','secret','url'];
        $encrypted_fields = ['secret'];

        // Set unset checkboxes
        $checkbox_fields = [];

        foreach ($checkbox_fields as $checkbox_field) {
            if (!isset($vars[$checkbox_field])) {
                $vars[$checkbox_field] = 'false';
            }
        }

        $this->Input->setRules($this->getRowRules($vars));

        // Validate module row
        if ($this->Input->validates($vars)) {
            // Build the meta data for this row
            $meta = [];
            foreach ($vars as $key => $value) {
                if (in_array($key, $meta_fields)) {
                    $meta[] = [
                        'key' => $key,
                        'value' => $value,
                        'encrypted' => in_array($key, $encrypted_fields) ? 1 : 0
                    ];
                }
            }

            return $meta;
        }
    }

    /**
     * {@inheritdoc}
     */
    public function editModuleRow($module_row, array &$vars)
    {
        $meta_fields = ['name','secret','url'];
        $encrypted_fields = ['secret'];

        // Set unset checkboxes
        $checkbox_fields = [];

        foreach ($checkbox_fields as $checkbox_field) {
            if (!isset($vars[$checkbox_field])) {
                $vars[$checkbox_field] = 'false';
            }
        }

        $this->Input->setRules($this->getRowRules($vars));

        // Validate module row
        if ($this->Input->validates($vars)) {
            // Build the meta data for this row
            $meta = [];
            foreach ($vars as $key => $value) {
                if (in_array($key, $meta_fields)) {
                    $meta[] = [
                        'key' => $key,
                        'value' => $value,
                        'encrypted' => in_array($key, $encrypted_fields) ? 1 : 0
                    ];
                }
            }

            return $meta;
        }
    }

    /**
     * {@inheritdoc}
     */
    private function getRowRules(&$vars)
    {
        $rules = [
            'name' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('Convoy.!error.name.valid', true)
                ]
            ],
            'secret' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('Convoy.!error.secret.valid', true)
                ],
            ],
            'url' => [
                'valid' => [
                    'rule' => [[$this, 'validateUrl']],
                    'message' => Language::_('Convoy.!error.url.valid', true)
                ],
                'validate_connection' => [
                    'rule' => [
                        [$this, 'validateConnection'],
                        $vars['secret']
                    ],
                    'message' => Language::_('Convoy.!error.validate_connection.valid', true)
                ]
            ]
        ];

        return $rules;
    }

    /**
     * Validates that the given url is valid.
     *
     * @param string $url The url to validate
     * @return bool True if the url is valid, false otherwise
     */
    public function validateUrl($url)
    {
        $validator = new Server();
        $url = str_replace(array('http://','https://'), '', $url);
        return $validator->isDomain($url) || $validator->isIp($url);
    }

    /**
     * Validates whether or not the connection details are valid by attempting to fetch data
     *
     * @param string $url The Convoy server url
     * @param string $secret The Convoy server secret
     * @return bool True if the connection is valid, false otherwise
     */
    public function validateConnection($url, $secret)
    {
        try {
            $api = $this->getApi($secret, $url);
            $response = $api->getNodes();

            $success = false;
            if ($response->status() == '200') {
                $success = true;
            }

            if ($success) {
                return true;
            }
        } catch (Exception $e) {
        }

        return false;
    }

    /**
     * {@inheritdoc}
     */
    public function getPackageFields($vars = null)
    {
        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();

        $module_row = $this->attemptGetModuleRow($vars);
        $api = $this->getApi($module_row->meta->secret, $module_row->meta->url);
        
        $all_nodes = $api->getAllNodes();
        //echo json_encode($all_nodes);
        $available_nodes = [];
        foreach ($all_nodes as $node) {
            $available_nodes[$node->id] = $node->name;
        }

        // Set the Nodes field
        $nodes = $fields->label(Language::_('Convoy.package_fields.nodes', true), 'convoy_nodes');
        $nodes->attach(
            $fields->fieldMultiSelect(
                'meta[nodes][]',
                $available_nodes,
                (isset($vars->meta['nodes']) ? $vars->meta['nodes'] : []),
                ['id' => 'convoy_nodes']
            )
        );
        $fields->setField($nodes);

        // Set the Template field
        $template = $fields->label(Language::_('Convoy.package_fields.template', true), 'convoy_template');
        $template->attach(
            $fields->fieldText(
                'meta[template]',
                (isset($vars->meta['template']) ? $vars->meta['template'] : null),
                ['id' => 'convoy_template']
            )
        );
        $template_tooltip = $fields->tooltip(Language::_('Convoy.package_fields.tooltip.template', true));
        $template->attach($template_tooltip);
        $fields->setField($template);

        // Set the CPU Limit field
        $cpu = $fields->label(Language::_('Convoy.package_fields.cpu', true), 'convoy_cpu');
        $cpu->attach(
            $fields->fieldNumber(
                'meta[cpu]',
                (isset($vars->meta['cpu']) ? $vars->meta['cpu'] : null),
                0,
                null,
                null,
                ['id' => 'convoy_cpu']
            )
        );
        $fields->setField($cpu);

        // Set the Memory (GB) field
        $memory = $fields->label(Language::_('Convoy.package_fields.memory', true), 'convoy_memory');
        $memory->attach(
            $fields->fieldNumber(
                'meta[memory]',
                (isset($vars->meta['memory']) ? $vars->meta['memory'] : null),
                0,
                null,
                0.001,
                ['id' => 'convoy_memory']
            )
        );
        $fields->setField($memory);

        // Set the Disk (GB) field
        $disk = $fields->label(Language::_('Convoy.package_fields.disk', true), 'convoy_disk');
        $disk->attach(
            $fields->fieldNumber(
                'meta[disk]',
                (isset($vars->meta['disk']) ? $vars->meta['disk'] : null),
                0,
                null,
                0.001,
                ['id' => 'convoy_disk']
            )
        );
        $fields->setField($disk);

        // Set the Bandwidth (GB) field
        $bandwidth = $fields->label(Language::_('Convoy.package_fields.bandwidth', true), 'convoy_bandwidth');
        $bandwidth->attach(
            $fields->fieldNumber(
                'meta[bandwidth]',
                (isset($vars->meta['bandwidth']) ? $vars->meta['bandwidth'] : null),
                0,
                null,
                0.001,
                ['id' => 'convoy_bandwidth']
            )
        );
        $fields->setField($bandwidth);

        // Set the Snapshots field
        $snapshots = $fields->label(Language::_('Convoy.package_fields.snapshots', true), 'convoy_snapshots');
        $snapshots->attach(
            $fields->fieldNumber(
                'meta[snapshots]',
                (isset($vars->meta['snapshots']) ? $vars->meta['snapshots'] : null),
                0,
                null,
                null,
                ['id' => 'convoy_snapshots']
            )
        );
        $fields->setField($snapshots);

        // Set the Backups field
        $backups = $fields->label(Language::_('Convoy.package_fields.backups', true), 'convoy_backups');
        $backups->attach(
            $fields->fieldNumber(
                'meta[backups]',
                (isset($vars->meta['backups']) ? $vars->meta['backups'] : null),
                0,
                null,
                null,
                ['id' => 'convoy_backups']
            )
        );
        $fields->setField($backups);

        // Set the IPv4 Count field
        $ipv4 = $fields->label(Language::_('Convoy.package_fields.ipv4', true), 'convoy_ipv4');
        $ipv4->attach(
            $fields->fieldNumber(
                'meta[ipv4]',
                (isset($vars->meta['ipv4']) ? $vars->meta['ipv4'] : null),
                0,
                null,
                null,
                ['id' => 'convoy_ipv4']
            )
        );
        $fields->setField($ipv4);

        // Set the IPv6 Count field
        $ipv6 = $fields->label(Language::_('Convoy.package_fields.ipv6', true), 'convoy_ipv6');
        $ipv6->attach(
            $fields->fieldNumber(
                'meta[ipv6]',
                (isset($vars->meta['ipv6']) ? $vars->meta['ipv6'] : null),
                0,
                null,
                null,
                ['id' => 'convoy_ipv6']
            )
        );
        $fields->setField($ipv6);

        return $fields;
    }

    /**
     * Retrieves a list of rules for validating adding/editing a package
     *
     * @param array $vars A list of input vars
     * @return array A list of rules
     */
    public function getPackageRules($vars = null)
    {
        $rules = [
            'meta[nodes]' => [
                'valid' => [
                    'rule' => [[$this, 'packageHasNodes']],
                    'message' => Language::_('Convoy.!error.meta[nodes].valid', true)
                ]
            ],
            'meta[template]' => [
                'valid' => [
                    'rule' => 'isEmpty',
                    'negate' => true,
                    'message' => Language::_('Convoy.!error.meta[template].valid', true)
                ]
            ],
            'meta[cpu]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+$/'],
                    'message' => Language::_('Convoy.!error.meta[cpu].valid', true)
                ]
            ],
            'meta[memory]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+((\.|\,)[0-9]{1,4})?$/'],
                    'message' => Language::_('Convoy.!error.meta[memory].valid', true)
                ]
            ],
            'meta[disk]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+((\.|\,)[0-9]{1,4})?$/'],
                    'message' => Language::_('Convoy.!error.meta[disk].valid', true)
                ]
            ],
            'meta[bandwidth]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+((\.|\,)[0-9]{1,4})?$/'],
                    'message' => Language::_('Convoy.!error.meta[bandwidth].valid', true)
                ]
            ],
            'meta[snapshots]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+$/'],
                    'message' => Language::_('Convoy.!error.meta[snapshots].valid', true)
                ]
            ],
            'meta[backups]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+$/'],
                    'message' => Language::_('Convoy.!error.meta[backups].valid', true)
                ]
            ],
            'meta[ipv4]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+$/'],
                    'message' => Language::_('Convoy.!error.meta[ipv4].valid', true)
                ]
            ],
            'meta[ipv6]' => [
                'valid' => [
                    'rule' => ['matches', '/^[0-9]+$/'],
                    'message' => Language::_('Convoy.!error.meta[ipv6].valid', true)
                ]
            ],
        ];

        return $rules;
    }

    /**
     * Validate that at least one node is selected on the package
     * 
     * @param array $nodes A list of nodes
     * @return bool True if at least one node was given, false otherwise
     */
    public function packageHasNodes($nodes)
    {
        return isset($nodes) ? strlen(json_encode($nodes)) > 0 : false;
    }

    /**
     * {@inheritdoc}
     */
    public function addService(
        $package,
        array $vars = null,
        $parent_package = null,
        $parent_service = null,
        $status = 'pending'
    ) {
        Loader::loadModels($this, ['ModuleClientMeta', 'Clients']);

        $row = $this->getModuleRow();
        if (!$row) {
            $this->Input->setErrors(
                ['module_row' => ['missing' => Language::_('Convoy.!error.module_row.missing', true)]]
            );
            return;
        }

        $module = $this->getModule();
        $client = $this->Clients->get($vars['client_id']);
        
        $this->validateService($package, $vars);

        if ($this->Input->errors()) {
            return;
        }

        // Set configurable options
        $package = $this->getConfigurableOptions($vars, $package);

        // Only provision the service if 'use_module' is true
        if ($vars['use_module'] == 'true') {
            $api = $this->getApi($row->meta->secret, $row->meta->url);

            $user_id = $this->ModuleClientMeta->get($vars['client_id'], 'convoy_user_id', $module->id);
            if ($user_id) {
                $user_id = $user_id->value;
            } else {
                $name = substr($client->first_name . ' ' . $client->last_name, 0, 40);
                $email = $client->email;
                $password = $this->generatePassword(24);

                $response = $api->createUser($name, $email, $password);
                if (substr($response->status(), 0, 1) != '2') {
                    $this->Input->setErrors(['api' => $response->errors()]);
                    return;
                }

                $data = $response->response()->data;
                $this->ModuleClientMeta->set(
                    $vars['client_id'],
                    $module->id,
                    0,
                    [
                        ['key' => 'convoy_user_id', 'value' => $data->id, 'encrypted' => 0],
                        ['key' => 'convoy_name', 'value' => $name, 'encrypted' => 0],
                        ['key' => 'convoy_email', 'value' => $email, 'encrypted' => 0],
                        ['key' => 'convoy_password', 'value' => $password, 'encrypted' => 1]
                    ]
                );
                $user_id = $data->id;
            }

            $node_id = $this->selectNode($api, $package->meta->nodes);
            $template_id = $this->selectTemplate($api, $node_id, $package->meta->template);
            $ip_address_ids = $this->selectIPAddresses($api, $node_id, $package->meta->ipv4, $package->meta->ipv6);

            $response = $api->createServer(
                $vars['name'], 
                $vars['hostname'],
                $vars['password'],
                $user_id,
                $node_id, 
                $template_id,
                [
                    'cpu' => $package->meta->cpu,
                    'memory' => $package->meta->memory * 1024 * 1024 * 1024,
                    'disk' => $package->meta->disk * 1024 * 1024 * 1024,
                    'bandwidth' => $package->meta->bandwidth * 1024 * 1024 * 1024 * 1024,
                    'snapshot' => $package->meta->snapshots,
                    'backup' => $package->meta->backups,
                ],
                $ip_address_ids
            );
            if (substr($response->status(), 0, 1) != '2') {
                $this->Input->setErrors(['api' => array_merge([$response->status()], $response->errors())]);
                return;
            }

            $vars['server_id'] = $response->response()->data->uuid;
        }

        // Return service fields
        return [
            [
                'key' => 'server_id',
                'value' => isset($vars['server_id']) ? $vars['server_id'] : null,
                'encrypted' => 0
            ],
            [
                'key' => 'name',
                'value' => $vars['name'],
                'encrypted' => 0
            ],
            [
                'key' => 'hostname',
                'value' => $vars['hostname'],
                'encrypted' => 0
            ],
            [
                'key' => 'password',
                'value' => $vars['password'],
                'encrypted' => 1
            ]
        ];
    }

    /**
     * {@inheritdoc}
     */
    public function editService($package, $service, array $vars = null, $parent_package = null, $parent_service = null)
    {
        $row = $this->getModuleRow();
        if (!$row) {
            $this->Input->setErrors(
                ['module_row' => ['missing' => Language::_('Convoy.!error.module_row.missing', true)]]
            );
            return;
        }

        $service_fields = $this->serviceFieldsToObject($service->fields);

        $this->validateServiceEdit($package, $vars, true);

        if ($this->Input->errors()) {
            return;
        }

        // Set configurable options
        $package = $this->getConfigurableOptions($vars, $package);

        // Only update the service if 'use_module' is true
        if ($vars['use_module'] == 'true') {
            $api = $this->getApi($row->meta->secret, $row->meta->url);

            $response = $api->updateServerInformation($vars['server_id'], $vars['name'], $vars['hostname']);
            if (substr($response->status(), 0, 1) != '2') {
                $this->Input->setErrors(['api' => [$response->status()]/*$response->errors()*/]);
                return;
            }

            // In theory this code should work
            // However the edit server function in Convoy v2 API...
            // ... doesn't quite wanna cooperate...
            // So this functionality will be included
            // once Convoy V3 is out

            /*$api = $this->getApi($row->meta->secret, $row->meta->url);

            $response = $api->getServerDetails($vars['server_id']);
            if (substr($response->status(), 0, 1) != '2') {
                $this->Input->setErrors(['api' => $response->errors()]);
                return;
            }

            $node_id = $response->response()->data->node_id;
            $address_ids = $this->getServerIPAddressIDs($api, $node_id, $vars['server_id'], $response->response()->data->limits->addresses);

            $v4_diff = $package->meta->ipv4 - count($address_ids[0]);
            $v6_diff = $package->meta->ipv6 - count($address_ids[1]);

            if ($v4_diff > 0) {
                $address_ids[0] = array_merge(
                    $address_ids[0],
                    $this->selectIPAddresses($api, $node_id, $v4_diff, 0)
                );
            } else if ($v4_diff < 0) {
                $to_remove = array_slice($address_ids[0], -abs($v4_diff), abs($v4_diff));
                $address_ids[0] = array_diff($address_ids[0], $to_remove);
            }

            if ($v6_diff > 0) {
                $address_ids[1] = array_merge(
                    $address_ids[1],
                    $this->selectIPAddresses($api, $node_id, 0, $v6_diff)
                );
            } else if ($v6_diff < 0) {
                $to_remove = array_slice($address_ids[1], -abs($v6_diff), abs($v6_diff));
                $address_ids[1] = array_diff($address_ids[1], $to_remove);
            }

            $response = $api->editServer(
                $vars['server_id'],
                [
                    'cpu' => $package->meta->cpu,
                    'memory' => $package->meta->memory * 1024 * 1024 * 1024,
                    'disk' => $package->meta->disk * 1024 * 1024 * 1024,
                    'bandwidth' => $package->meta->bandwidth * 1024 * 1024 * 1024 * 1024,
                    'snapshot' => $package->meta->snapshots,
                    'backup' => $package->meta->backups,
                ],
                array_merge($address_ids[0], $address_ids[1])
            );*/
        }

        // Return all the service fields
        $encrypted_fields = ['password'];
        $return = [];
        $fields = ['server_id','name','hostname','password'];
        foreach ($fields as $field) {
            if (isset($vars[$field]) || isset($service_fields->{$field})) {
                $return[] = [
                    'key' => $field,
                    'value' => $vars[$field] ?? $service_fields->{$field},
                    'encrypted' => (in_array($field, $encrypted_fields) ? 1 : 0)
                ];
            }
        }

        return $return;
    }

    /**
     * {@inheritdoc}
     */
    public function suspendService($package, $service, $parent_package = null, $parent_service = null)
    {
        if (($row = $this->getModuleRow())) {
            $api = $this->getApi($row->meta->secret, $row->meta->url);

            $service_fields = $this->serviceFieldsToObject($service->fields);
            $response = $api->suspendServer($service_fields->server_id);

            if (substr($response->status(), 0, 1) != '2') {
                $this->Input->setErrors(['api' => array_merge([$response->status()],[$response->raw()],$response->errors())]);
                return;
            }
        }

        return null;
    }

    /**
     * {@inheritdoc}
     */
    public function unsuspendService($package, $service, $parent_package = null, $parent_service = null)
    {
        if (($row = $this->getModuleRow())) {
            $api = $this->getApi($row->meta->secret, $row->meta->url);

            $service_fields = $this->serviceFieldsToObject($service->fields);
            $response = $api->unsuspendServer($service_fields->server_id);

            if (substr($response->status(), 0, 1) != '2') {
                $this->Input->setErrors(['api' => $response->errors()]);
                return;
            }
        }

        return null;
    }

    /**
     * {@inheritdoc}
     */
    public function cancelService($package, $service, $parent_package = null, $parent_service = null)
    {
        if (($row = $this->getModuleRow())) {
            $api = $this->getApi($row->meta->secret, $row->meta->url);

            $service_fields = $this->serviceFieldsToObject($service->fields);
            $response = $api->deleteServer($service_fields->server_id);

            if (substr($response->status(), 0, 1) != '2') {
                $this->Input->setErrors(['api' => $response->errors()]);
                return;
            }
        }

        return null;
    }

    /**
     * {@inheritdoc}
     */
    public function renewService($package, $service, $parent_package = null, $parent_service = null)
    {
        return null;
    }

    /**
     * {@inheritdoc}
     */
    public function validateService($package, array $vars = null)
    {
        $this->Input->setRules($this->getServiceRules($vars));
        return $this->Input->validates($vars);
    }

    /**
     * {@inheritdoc}
     */
    public function validateServiceEdit($service, array $vars = null)
    {
        $this->Input->setRules($this->getServiceRules($vars, true));
        return $this->Input->validates($vars);
    }

    /**
     * {@inheritdoc}
     */
    private function getServiceRules(array $vars = null, $edit = false)
    {
        // Validate the service fields
        $rules = [
            'server_id' => [
                'valid' => [
                    'rule' => true,
                    'message' => Language::_('Convoy.!error.server_id.valid', true)
                ]
            ],
            'name' => [
                'valid' => [
                    'rule' => 'isEmpty',
                    'negate' => true,
                    'message' => Language::_('Convoy.!error.name.valid', true)
                ]
            ],
            'hostname' => [
                'valid' => [
                    'rule' => 'isEmpty',
                    'negate' => true,
                    'message' => Language::_('Convoy.!error.name.valid', true)
                ]
            ],
            'password' => [
                'valid' => [
                    'rule' => ['matches', '/^(?=.*[a-z])(?=.*[A-Z])(?=.*[0-9])(?=.*[!@#\$%\^&\*])(?=.{8,})/'],
                    'message' => Language::_('Convoy.!error.password.valid', true)
                ]
            ]
        ];

        if ($edit) {
            unset($rules['password']);
        }

        return $rules;
    }

    /**
     * {@inheritdoc}
     */
    public function changeServicePackage(
        $package_from,
        $package_to,
        $service,
        $parent_package = null,
        $parent_service = null
    ) {
        if (($row = $this->getModuleRow())) {
////            $api = $this->getApi(
////                $row->meta->host_name,
////                $row->meta->user_name,
////                $row->meta->password,
////                ($row->meta->use_ssl == 'true')
////            );
////
        }

        return null;
    }

    /**
     * {@inheritdoc}
     */
    public function addPackage($vars = null)
    {
        $this->Input->setRules($this->getPackageRules($vars));

        $meta = [];
        if ($this->Input->validates($vars)) {
            // Return all package meta fields
            foreach ($vars['meta'] as $key => $value) {
                $meta[] = [
                    'key' => $key,
                    'value' => $value,
                    'encrypted' => 0
                ];
            }
        }

        return $meta;
    }
    
    /**
     * {@inheritdoc}
     */
    public function editPackage($package, array $vars = null)
    {
        $this->Input->setRules($this->getPackageRules($vars));

        $meta = [];
        if ($this->Input->validates($vars)) {
            // Return all package meta fields
            foreach ($vars['meta'] as $key => $value) {
                $meta[] = [
                    'key' => $key,
                    'value' => $value,
                    'encrypted' => 0
                ];
            }
        }

        return $meta;
    }

    /**
     * Initializes the ConvoyApi and returns an instance of that object.
     *
     * @param string $secret Placeholder description
     * @param string $url Placeholder description
     * @return ConvoyApi The ConvoyApi instance
     */
    private function getApi($secret, $url)
    {
        Loader::load(dirname(__FILE__) . DS . 'apis' . DS . 'convoy_api.php');
        $api = new ConvoyApi($secret, $url);
        return $api;
    }
  
    /**
     * Generates a password.
     *
     * @param int $min_length The minimum character length for the password (10 or larger)
     * @param int $max_length The maximum character length for the password (256 or fewer)
     * @return string The generated password
     */
    private function generatePassword($length)
    {
        if($length < 10)
            $length = 10;
    
        $sets = array();
        $sets[] = 'ABCDEFGHJKLMNPQRSTUVWXYZ';
        $sets[] = 'abcdefghjkmnpqrstuvwxyz';
        $sets[] = '23456789';
        $sets[]  = '~!@#$%^&*(){}[],./?';
    
        $password = '';
        
        foreach ($sets as $set) {
            $password .= $set[array_rand(str_split($set))];
        }
    
        while(strlen($password) < $length) {
            $randomSet = $sets[array_rand($sets)];
            $password .= $randomSet[array_rand(str_split($randomSet))]; 
        }
        
        return str_shuffle($password);
    }

    /**
     * {@inheritdoc}
     */
    public function getAdminServiceInfo($service, $package)
    {
        Loader::loadModels($this, ['ModuleClientMeta']);
        $row = $this->getModuleRow();
        $api = $this->getApi($row->meta->secret, $row->meta->url);
        $service_fields = $this->serviceFieldsToObject($service->fields);

        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('admin_service_info', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'convoy' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html']);

        $module = $this->getModule();
        $user_id = $this->ModuleClientMeta->get($service->client_id, 'convoy_user_id', $module->id)->value;
        $email = $this->ModuleClientMeta->get($service->client_id, 'convoy_email', $module->id)->value;
        $password = $this->ModuleClientMeta->get($service->client_id, 'convoy_password', $module->id)->value;
        $sso_response = $api->createSSOToken($user_id);
        $server_response = $api->getServer($service_fields->server_id);

        $this->view->set('module_row', $row);
        $this->view->set('package', $package);
        $this->view->set('service', $service);
        $this->view->set('service_fields', $service_fields);
        $this->view->set('email', $email);
        $this->view->set('password', $password);
        if (substr($sso_response->status(), 0, 1) == '2') {
            $this->view->set('login_url', $row->meta->url . '/authenticate?token=' . $sso_response->response()->data->token);
        }
        if (substr($server_response->status(), 0, 1) == '2') {
            $response = $server_response->response();
            $this->view->set('ip_addresses', array_map(
                function($ip) {
                    return $ip->address;
                }, 
                $response->data->limits->addresses->ipv4, 
                $response->data->limits->addresses->ipv6
            ));
        }

        return $this->view->fetch();
    }

    /**
     * {@inheritdoc}
     */
    public function getClientServiceInfo($service, $package)
    {
        Loader::loadModels($this, ['ModuleClientMeta']);
        $row = $this->getModuleRow();
        $api = $this->getApi($row->meta->secret, $row->meta->url);
        $service_fields = $this->serviceFieldsToObject($service->fields);

        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('client_service_info', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'convoy' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, ['Form', 'Html']);

        $module = $this->getModule();
        $user_id = $this->ModuleClientMeta->get($service->client_id, 'convoy_user_id', $module->id)->value;
        $email = $this->ModuleClientMeta->get($service->client_id, 'convoy_email', $module->id)->value;
        $password = $this->ModuleClientMeta->get($service->client_id, 'convoy_password', $module->id)->value;
        $sso_response = $api->createSSOToken($user_id);
        $server_response = $api->getServer($service_fields->server_id);

        $this->view->set('module_row', $row);
        $this->view->set('package', $package);
        $this->view->set('service', $service);
        $this->view->set('service_fields', $service_fields);
        $this->view->set('email', $email);
        $this->view->set('password', $password);
        if (substr($sso_response->status(), 0, 1) == '2') {
            $this->view->set('login_url', $row->meta->url . '/authenticate?token=' . $sso_response->response()->data->token);
        }
        if (substr($server_response->status(), 0, 1) == '2') {
            $response = $server_response->response();
            $this->view->set('ip_addresses', array_map(
                function($ip) {
                    return $ip->address;
                }, 
                $response->data->limits->addresses->ipv4, 
                $response->data->limits->addresses->ipv6
            ));
        }

        return $this->view->fetch();
    }

    /**
     * {@inheritdoc}
     */
    public function getAdminAddFields($package, $vars = null)
    {
        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();

        // Set the Server ID field
        $server_id = $fields->label(Language::_('Convoy.service_fields.server_id', true), 'convoy_server_id');
        $server_id->attach(
            $fields->fieldText(
                'server_id',
                (isset($vars->server_id) ? $vars->server_id : null),
                ['id' => 'convoy_server_id']
            )
        );
        $fields->setField($server_id);

        // Set the Name field
        $name = $fields->label(Language::_('Convoy.service_fields.name', true), 'convoy_name');
        $name->attach(
            $fields->fieldText(
                'name',
                (isset($vars->name) ? $vars->name : null),
                ['id' => 'convoy_name']
            )
        );
        $fields->setField($name);

        // Set the Hostname field
        $hostname = $fields->label(Language::_('Convoy.service_fields.hostname', true), 'convoy_hostname');
        $hostname->attach(
            $fields->fieldText(
                'hostname',
                (isset($vars->hostname) ? $vars->hostname : null),
                ['id' => 'convoy_hostname']
            )
        );
        $fields->setField($hostname);

        // Set the Password field
        $password = $fields->label(Language::_('Convoy.service_fields.password', true), 'convoy_password');
        $password->attach(
            $fields->fieldPassword(
                'password',
                ['id' => 'convoy_password', 'value' => (isset($vars->password) ? $vars->password : null)]
            )
        );
        $fields->setField($password);

        return $fields;
    }

    /**
     * {@inheritdoc}
     */
    public function getAdminEditFields($package, $vars = null)
    {
        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();

        // Set the Server ID field
        $server_id = $fields->label(Language::_('Convoy.service_fields.server_id', true), 'convoy_server_id');
        $server_id->attach(
            $fields->fieldText(
                'server_id',
                (isset($vars->server_id) ? $vars->server_id : null),
                ['id' => 'convoy_server_id']
            )
        );
        $fields->setField($server_id);

        // Set the Name field
        $name = $fields->label(Language::_('Convoy.service_fields.name', true), 'convoy_name');
        $name->attach(
            $fields->fieldText(
                'name',
                (isset($vars->name) ? $vars->name : null),
                ['id' => 'convoy_name']
            )
        );
        $fields->setField($name);

        // Set the Hostname field
        $hostname = $fields->label(Language::_('Convoy.service_fields.hostname', true), 'convoy_hostname');
        $hostname->attach(
            $fields->fieldText(
                'hostname',
                (isset($vars->hostname) ? $vars->hostname : null),
                ['id' => 'convoy_hostname']
            )
        );
        $fields->setField($hostname);

        return $fields;
    }

    /**
     * {@inheritdoc}
     */
    public function getClientAddFields($package, $vars = null)
    {
        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();

        // Set the Name field
        $name = $fields->label(Language::_('Convoy.service_fields.name', true), 'convoy_name');
        $name->attach(
            $fields->fieldText(
                'name',
                (isset($vars->name) ? $vars->name : null),
                ['id' => 'convoy_name']
            )
        );
        $fields->setField($name);

        // Set the Hostname field
        $hostname = $fields->label(Language::_('Convoy.service_fields.hostname', true), 'convoy_hostname');
        $hostname->attach(
            $fields->fieldText(
                'hostname',
                (isset($vars->hostname) ? $vars->hostname : null),
                ['id' => 'convoy_hostname']
            )
        );
        $fields->setField($hostname);

        // Set the Password field
        $password = $fields->label(Language::_('Convoy.service_fields.password', true), 'convoy_password');
        $password->attach(
            $fields->fieldPassword(
                'password',
                ['id' => 'convoy_password', 'value' => (isset($vars->password) ? $vars->password : null)]
            )
        );
        $fields->setField($password);

        return $fields;
    }

    /**
     * {@inheritdoc}
     */
    public function getClientEditFields($package, $vars = null)
    {
        Loader::loadHelpers($this, ['Html']);

        $fields = new ModuleFields();

        // Set the Name field
        $name = $fields->label(Language::_('Convoy.service_fields.name', true), 'convoy_name');
        $name->attach(
            $fields->fieldText(
                'name',
                (isset($vars->name) ? $vars->name : null),
                ['id' => 'convoy_name']
            )
        );
        $fields->setField($name);

        // Set the Hostname field
        $hostname = $fields->label(Language::_('Convoy.service_fields.hostname', true), 'convoy_hostname');
        $hostname->attach(
            $fields->fieldText(
                'hostname',
                (isset($vars->hostname) ? $vars->hostname : null),
                ['id' => 'convoy_hostname']
            )
        );
        $fields->setField($hostname);

        return $fields;
    }

    /** 
     * Attempt to get a module row from $vars included in some functions
     * 
     * @param mixed $vars stdClass or array of vars
     * @return $stdClass a stdClass representing the module row
     */
    private function attemptGetModuleRow($vars)
    {
        if (is_array($vars)) {
            $vars = (object)$vars;
        }

        $module_row = null;
        if (isset($vars->module_group) && $vars->module_group == '' || $vars->module_group == 'select') {
            if (isset($vars->module_row) && $vars->module_row > 0) {
                $module_row = $this->getModuleRow($vars->module_row);
            } else {
                $rows = $this->getModuleRows();
                if (isset($rows[0])) {
                    $module_row = $rows[0];
                }
                unset($rows);
            }
        } else {
            $rows = $this->getModuleRows($vars->module_group);
            if (isset($rows[0])) {
                $module_row = $rows[0];
            }
            unset($rows);
        }

        return $module_row;
    }

    /**
     * Returns an array of package fields, overriding the configurable options
     *
     * @param array $vars An array of key/value input pairs
     * @param stdClass $package A stdClass object representing the package for the service
     * @return stdClass The modified package object
     */
    private function getConfigurableOptions(array $vars, $package)
    {
        $fields = ['nodes', 'template', 'cpu', 'memory', 'disk', 'bandwidth', 'snapshots', 'backups', 'ipv4', 'ipv6'];

        // Override package fields, if an equivalent configurable option exists
        if (!empty($vars['configoptions'])) {
            foreach ($vars['configoptions'] as $field => $value) {
                if (in_array($field, $fields)) {
                    $package->meta->{$field} = $value;
                }
            }
        }

        return $package;
    }

    /**
     * Selects a node with least RAM in use
     * 
     * @param ConvoyApi $api Instance of the Convoy API
     * @param array $node_ids Array of available node ids
     * @return int ID of the least occupied node, -1 if errors occured
     */
    private function selectNode($api, $node_ids)
    {
        if (count($node_ids) == 1) {
            return $node_ids[0];
        }

        $nodes = [];
        foreach ($node_ids as $node_id) {
            $response = $api->getNode($node_id);
            if ($response->status() != '200') {
                continue;
            }

            $node = $response->response()->data;

            // Apply overallocation
            $node->memory = $node->memory * ((100 + $node->memory_overallocate) / 100); 
            $node->disk = $node->disk * ((100 + $node->disk_overallocate) / 100);

            // Calcualte free resources
            $node->memory_free = $node->memory - $node->memory_allocated;
            $node->disk_free = $node->disk - $node->disk_allocated;

            $nodes[$node_id] = $node;
        }

        $max_free_memory = -1;  // Initialize with a negative value
        $max_free_memory_node_id = $node_ids[0];

        foreach ($nodes as $id => $node) {
            if ($node->memory_free > $max_free_memory) {
                $max_free_memory = $node->memory_free;
                $max_free_memory_node_id = $id;
            }
        }        
        
        return $max_free_memory_node_id;
    }

    /**
     * Selects a matching template from the specified node
     * 
     * @param ConvoyApi $api Instance of the Convoy API
     * @param int $node_id ID of the node to search templates on
     * @param string $template Partial template name to find a match with
     * @return int ID of the first matched template, -1 if errors occured or not found
     */
    private function selectTemplate($api, $node_id, $template)
    {
        $response = $api->getTemplates($node_id);
        if (substr($response->status(), 0, 1) != '2') {
            return -1;
        }

        $templates = $response->response()->data;
        foreach ($templates as $template_group) {
            if (str_contains(strtolower($template_group->name), strtolower($template)) &&
                isset($template_group->templates->data[0])) {
                return $template_group->templates->data[0]->uuid;
            }

            foreach ($template_group->templates->data as $node_template) {
                if (str_contains(strtolower($node_template->name), strtolower($template))) {
                    return $node_template->uuid;
                }
            }
        }

        return -1;
    }

    /**
     * Selects the specified amount of the available v4 and v6 IP addresses from the node
     * 
     * @param ConvoyApi $api Instance of the Convoy API
     * @param int $node_id ID of the node to select IPs from
     * @param int $v4_count Amount of V4 addresses to select
     * @param int $v6_count Amount of V6 addresses to select
     * @return array Two element array, where index 0 is IPv4 addresse and index 1 is IPv6 addresse IDs, null if errors occured
     */
    private function selectIPAddresses($api, $node_id, $v4_count, $v6_count)
    {
        $response = $api->getAddresses($node_id);
        if (substr($response->status(), 0, 1) != '2') {
            return null;
        }

        $addresses = [];
        $selected_v4 = 0;
        $selected_v6 = 0;

        $available_addresses = $response->response()->data;
        foreach ($available_addresses as $address) {
            if ($address->server_id != null) {
                continue;
            }

            if ($address->type == 'ipv4') {
                if ($selected_v4 < $v4_count) {
                    $addresses[] = $address->id;
                    $selected_v4++;
                } 
            } else if ($selected_v6 < $v6_count) {
                $addresses[] = $address->id;
                $selected_v6++;
            }
        }

        return $addresses;
    }
}